/* goGame.c -- the go game interface */

/* mango/tango -- copyright 1994 Edward Grigassy
 * the code included in this package may be distributed, read, and copied
 * freely BUT NOT SOLD OR INCLUDED IN OTHER SOFTWARE, for commercial or
 * public/free use, without express written consent of the author.
 */

#include "public.h"
#include "strmisc.h"

#include "goGameP.h"

/* this array must correspond to the PLAYER_ defines */
char *playerNames[] = {
   "unknown",
   "black",
   "white",
   "observer",
   NULL
};

/* ------------------------- goGame functions ------------------------- */

int boardGetSpot(goGame *theGame, int i, int j)
{
   char spot = theGame->state[(j*theGame->numRows)+i];
   int ret;

   ret = (spot == BOARD_WHITE? PLAYER_WHITE:
	  (spot == BOARD_BLACK? PLAYER_BLACK: PLAYER_NONE));
   return ret;
}

int boardSetSpot(goGame *theGame, int i, int j, int val)
{
   char spot = theGame->state[(j*theGame->numRows)+i];

   if (spot == BOARD_BLACK) theGame->blackCount--;
   else if (spot == BOARD_WHITE) theGame->whiteCount--;
   if (val == PLAYER_WHITE)
   {
      spot = BOARD_WHITE;
      theGame->whiteCount++;
   }
   else if (val == PLAYER_BLACK)
   {
      spot = BOARD_BLACK;
      theGame->blackCount++;
   }
   else
      spot = BOARD_EMPTY;
   theGame->state[(j*theGame->numRows)+i] = spot;
   return val;
}

int isStarPoint(goGame *theGame, int i, int j)
{
   int midRow, midColumn, starOffset, ret;
   
   midRow = (theGame->numRows)/2;
   midColumn = (theGame->numColumns)/2;
   starOffset = theGame->starPoint-1;
   ret = ((i == starOffset || i == midColumn ||
	   i == theGame->numColumns-starOffset-1) &&
	  (j == starOffset || j == midColumn ||
	   j == theGame->numRows-starOffset-1));
   return ret;
}

void tallyStoneCounts(char *board, int *whites, int *blacks)
{
   if (!strEmpty(board) && whites && blacks)
   {
      char *at;

      *whites = 0;
      *blacks = 0;
      for (at = board; *at; at++)
      {
	 if (*at == BOARD_WHITE) (*whites)++;
	 else if (*at == BOARD_BLACK) (*blacks)++;
      }
   }
}

void outputBoardText(FILE *out, char *board, int rows, int columns)
{
   int j;
   char area[BOARD_MAX_LEN];
   
   for (j = 0; j < rows; j++)
   {
      strncpy(area, &(board[j*columns]), columns);
      area[columns] = '\0';
      fprintf(out, "%s\n", area);
   }
}

goGame *goGameCreate(int rows, int columns, int starPoint, int turn)
{
   goGame *ret = NEW(goGame);
   if (ret)
   {
      ret->numRows = rows;
      ret->numColumns = columns;
      ret->starPoint = starPoint;
      ret->turn = turn;
      ret->keepHistory = True;
      ret->playLevel = PLAY_HANDICAP;
      /* empty the board */
      goGameClear(ret);
   }

   return ret;
}

goGame *goGameClear(goGame *theGame)
{
   if (theGame)
   {
      int i, numCells = theGame->numRows * theGame->numColumns;

      theGame->numPasses = 0;
      for (i = 0; i < numCells; i++)
	 theGame->state[i] = BOARD_EMPTY;
      theGame->state[i] = '\0';
      goGameCleanHistory(theGame);
   }
   return theGame;
}

goGame *goGameCopy(goGame *orig, int keepHistory)
{
   goGame *ret = NULL;
   if (orig)
      ret = NEW(goGame);
   if (ret)
   {
      ret->userOptions = orig->userOptions;
      ret->numRows = orig->numRows;
      ret->numColumns = orig->numColumns;
      ret->starPoint = orig->starPoint;
      ret->me = orig->me;
      ret->turn = orig->turn;
      ret->whiteCount = orig->whiteCount;
      ret->blackCount = orig->blackCount;
      strcpy(ret->state, orig->state);
      ret->myId = orig->myId;
      ret->keepHistory = keepHistory;
      
      /* create the copy of history if the user wants it */
      if (keepHistory)
      {
	 goHistory *parse;
	 
	 for (parse = orig->historyFirst; parse; parse = parse->next)
	 {
	    goMessage *new = goMessageCopy(parse->turn);
	    goGameAddToHistory(ret, new);
	 }
      }
   }
   return ret;
}

/* returns 0 if game boards are equivalent,
 * != 0 if game boards different
 */

int goGameCompare(goGame *l, goGame *r)
{
   int compare;

   compare = l->whiteCount - r->whiteCount;
   if (compare == 0)
      compare = l->blackCount - r->blackCount;
   if (compare == 0)
      compare = strcmp(l->state, r->state);
   return compare;
}

int goGameSetKeepHistory(goGame *theGame, int historyOn)
{
   int ret = False;

   if (theGame)
   {
      theGame->keepHistory = (Boolean)historyOn;
   }
   return ret;
}

int goGameGetKeepHistory(goGame *theGame)
{
   int ret = False;

   if (theGame)
   {
      ret = theGame->keepHistory;
   }
   return ret;
}

int goGameGetNumPasses(goGame *theGame)
{
   int ret = 0;
   if (theGame) ret = theGame->numPasses;
   return ret;
}

int goGameSetId(goGame *theGame, int myId)
{
   int ret = -1;
   if (theGame)
      ret = theGame->myId = myId;
   return ret;
}

int goGameGetId(goGame *theGame)
{
   int ret = -1;
   if (theGame) ret = theGame->myId;
   return ret;
}

char *goGameBoardString(goGame *theGame)
{
   static char retarea[1024];
   
   strcpy(retarea,"");
   if (theGame && theGame->state)
   {
      strcpy(retarea, theGame->state);
   }
   
   return retarea;
}

goGame *goGameCleanHistory(goGame *theGame)
{
   if (theGame)
   {
      goHistory *bye, *temp;
	 
      /* free all the history */
      for (bye = theGame->historyFirst; bye; )
      {
	 temp = bye->next;
	 goHistoryDestroy(bye);
	 bye = temp;
      }
      theGame->historyFirst = NULL;
      theGame->historyLast = NULL;
   }

   return theGame;
}


goGame *goGameDestroy(goGame *theGame)
{
   if (theGame)
   {
      if (theGame->historyFirst)
      {
	 goGameCleanHistory(theGame);
      }
      DESTROY(theGame);
   }
   return NULL;
}

int goGameSetOption(goGame *theGame, long int option, int high)
{
   int ret = 0;

   if (theGame && option)
   {
      if (high) /* set true */
	 theGame->userOptions = theGame->userOptions | option;
      else /* set false */
	 theGame->userOptions = theGame->userOptions & ~option;
      ret = 1;
   }
   
   return ret;
}

int goGameGetOption(goGame *theGame, long int option)
{
   int ret = 0;

   if (theGame && option)
   {
      ret = (theGame->userOptions & option? 1: 0);
   }
   
   return ret;
}

int goGameSave(goGame *theGame, char *saveFile)
{
   FILE *sf = NULL;
   int saved = False;

   strTrimEnd(saveFile);
   if (theGame && saveFile && (sf = fopen(saveFile,"w")))
   {
      goHistory *scan;
      
      fprintf(sf, "userOptions: %lx\n", theGame->userOptions);
      fprintf(sf, "numRows: %d\n", theGame->numRows);
      fprintf(sf, "numColumns: %d\n", theGame->numColumns);
      fprintf(sf, "starPoint: %d\n", theGame->starPoint);
      fprintf(sf, "turn: %d\n", theGame->turn);
      /* saving "me" is not important, let the user remember */
      fprintf(sf, "currentBoard:\n");
      outputBoardText(sf, theGame->state,
		      theGame->numRows, theGame->numColumns);
      fprintf(sf, "beginHistory:\n");
      for (scan = theGame->historyFirst; scan; scan = scan->next)
      {
	 goMessageSave(sf, scan->turn);
      }
      fprintf(sf, "endHistory.\n");
      fclose(sf);
      saved = True;
   }
   return saved;
}

int goGameSetPlayer(goGame *theGame, int player)
{
   int ret = -1;
   if (theGame) ret = theGame->me = player;
   return ret;
}

int goGameGetPlayer(goGame *theGame)
{
   int ret = -1;
   if (theGame) ret = theGame->me;
   return ret;
}

int goGameGetTurn(goGame *theGame)
{
   int ret = -1;
   if (theGame) ret = theGame->turn;
   return ret;
}

int goGameSetTurn(goGame *theGame, int player)
{
   int ret = -1;
   if (theGame) ret = theGame->turn = player;
   return ret;
}

int goGameLoad(goGame *theGame, char *loadFile)
{
   FILE *lf = NULL;
   int loaded = False;

   if (theGame && loadFile && (lf = fopen(loadFile,"r")))
   {
      char line[256];
      goMessage *mess = goMessageCreate();

      /* the interface update procedure will not receive individual
       * message events of the load */
      goMessageFill(mess, theGame, MESSAGE_NEW_GAME, 0, 0, theGame->me, NULL);
      goMessageApply(theGame, mess);
      fscanf(lf, "%*s %lx\n", &theGame->userOptions);
      fscanf(lf, "%*s %d\n", &theGame->numRows);
      fscanf(lf, "%*s %d\n", &theGame->numColumns);
      fscanf(lf, "%*s %d\n", &theGame->starPoint);
      fscanf(lf, "%*s %d\n", &theGame->turn);
      /* loading "me" is not important, let the user remember */
      /* should I load in the board or just the history? */
      /* for now just read the history ... */
      while (fgets(line, 255, lf) && !strstr(line, "beginHistory"));
      while (fgets(line, 255, lf) && !strstr(line, "endHistory"))
      {
	 int type, player, x, y, numRows, numColumns;
	 sscanf(line,"history: %d, %d, %d, %d", &type, &player, &x, &y);
	 goMessageFill(mess, theGame, type, x, y, player, NULL);
	 mess->override = (char)True;
	 goMessageApply(theGame, mess);
      }
      goMessageDestroy(mess);
      fclose(lf);
      loaded = True;
   }
   return loaded;
}

/* -------------------------- goHistory functions ------------------------- */

goHistory *goHistoryDestroy(goHistory *bye)
{
   if (bye)
   {
      goMessageDestroy(bye->turn);
      DESTROY(bye);
      bye = NULL;
   }
   return bye;
}

/* keeps message around (so don't free this memory!)...
 * also: current philosophy is to keep record of the board
 * (in the currentBoard field of the goMessage, if field used)
 * BEFORE the recorded event takes effect */

goGame *goGameAddToHistory(goGame *theGame, goMessage *message)
{
   goGame *ret = NULL;
   
   if (theGame && message)
   {
      goHistory *node = NEW(goHistory);

      if (node)
      {
	 node->turn = message;
	 strcpy(node->board, theGame->state);
	 tallyStoneCounts(node->board,
			  &(node->whiteCount), &(node->blackCount));
	 node->next = NULL;
	 node->prev = NULL;
	 if (theGame->historyFirst == NULL &&
	     theGame->historyLast == NULL)
	 {
	    /* no history yet */
	    theGame->historyFirst = node;
	    theGame->historyLast = node;
	 }
	 else
	 {
	    /* we're dealing with 1 or more items of history, insert at end */
	    theGame->historyLast->next = node;
	    node->prev = theGame->historyLast;
	    theGame->historyLast = node;
	 }
	 ret = theGame;
      }
   }
   return ret;
}

goHistory *lastHist = NULL;

goMessage *goGameGetMessage(goGame *game)
{
   goMessage *ret = NULL;

   if (game)
   {
      lastHist = game->historyFirst;
   }
   else if (lastHist)
   {
      lastHist = lastHist->next;
   }
   if (lastHist) ret = lastHist->turn;

   return ret;
}

goHistory *goHistorySearchLast(goHistory *from, int messageType)
{
   if (from)
   {
      while (from && from->turn && from->turn->type != messageType)
	 from = from->prev;
   }
   
   return from;
}

/* find the last event corresponding to message and remove it */

goGame *goGameRemoveFromHistory(goGame *theGame, goMessage *message)
{
   goGame *ret = NULL;
   
   if (theGame && message)
   {
      goHistory *search = theGame->historyLast;

      while (search && goMessageCompare(search->turn, message) != 0)
	 search = goHistorySearchLast(search, message->type);
      if (search)
      {
	 if (search->prev && search->next)
	 {
	    /* somewhere in the middle of history */
	    search->prev->next = search->next;
	    search->next->prev = search->prev;
	 }
	 else if (search->next)
	 {
	    /* at beginning of history */
	    search->next->prev = NULL;
	    theGame->historyFirst = search->next;
	 }
	 else if (search->prev)
	 {
	    /* at end of history */
	    search->prev->next = NULL;
	    theGame->historyLast = search->prev;
	 }
	 else
	 {
	    /* this was the only item of history */
	    theGame->historyFirst = NULL;
	    theGame->historyLast = NULL;
	 }
	 goHistoryDestroy(search);
	 ret = theGame;
      }
   }
   return ret;
}

/*-------------------------- goMessage functions --------------------------- */

goMessage *goMessageCreate(void)
{
   goMessage *ret = NEW(goMessage);

   return ret;
}

goMessage *goMessageDestroy(goMessage *mess)
{
   if (mess)
   {
      DESTROY(mess);
      mess = NULL;
   }
   return mess;
}

char *goMessageTypeStr(int type)
{
   char *ret = "(Undefined)";

   switch(type)
   {
    case MESSAGE_COMMENT:	ret = "COMMENT"; break;
    case MESSAGE_PLACE_STONE:	ret = "PLACE_STONE"; break;
    case MESSAGE_CAPTURE_STONE:	ret = "CAPTURE_STONE"; break;
    case MESSAGE_CAPTURE_GROUP:	ret = "CAPTURE_GROUP"; break;
    case MESSAGE_UNDO_MOVE:	ret = "UNDO_MOVE"; break;
    case MESSAGE_NOTIFY_SERVER:	ret = "NOTIFY_SERVER"; break;
    case MESSAGE_NOTIFY_CLIENT:	ret = "NOTIFY_CLIENT"; break;
    case MESSAGE_NEW_GAME:	ret = "NEW_GAME"; break;
    case MESSAGE_TRANSFER_HIST:	ret = "TRANSFER_HIST"; break;
    case MESSAGE_TRANSFER_DONE:	ret = "TRANSFER_DONE"; break;
    case MESSAGE_REQUEST_MOVE:	ret = "REQUEST_MOVE"; break;
    case MESSAGE_REQUEST_UNDO:	ret = "REQUEST_UNDO"; break;
    case MESSAGE_PASS:		ret = "PASS"; break;
    case MESSAGE_DISCONNECT:	ret = "DISCONNECT"; break;
   }
   return ret;
}

char *goMessageStr(goMessage *mess)
{
   static char retarea[512];

   sprintf(retarea,"owner %hx, %s, %s(%d, %d), player %d, [%s]",
	   mess->owner, goMessageTypeStr(mess->type),
	   (mess->override? "OVR, ": ""),
	   mess->x, mess->y, mess->player,
	   mess->comment);
   
   return retarea;
}

goMessage *goMessageCopy(goMessage *orig)
{
   goMessage *ret = NULL;

   if (orig)
   {
      ret = goMessageCreate();
      if (ret)
      {
	 bcopy(orig, ret, sizeof(goMessage));
      }
   }
   return ret;
}

void goMessageSave(FILE *out, goMessage *mess)
{
   if (out && mess)
   {
      fprintf(out,"history: %d, %d, %d, %d\n",
	      mess->type, mess->player, mess->x, mess->y);

      /* save current board? */
   }
}

goMessage *goMessageFill(goMessage *fillMe, goGame *theGame,
			 int messageType, int x, int y, int player,
			 char *comment)
{
   goMessage *ret = NULL;
   
   if (fillMe && theGame)
   {
      fillMe->type = messageType;
      fillMe->owner = theGame->myId;
      fillMe->player = player;
      fillMe->x = x;
      fillMe->y = y;
      fillMe->numRows = theGame->numRows;
      fillMe->numColumns = theGame->numColumns;
      fillMe->override = False;
      if (comment)
      {
	 strncpy(fillMe->comment, comment, COMMENT_LEN);
	 fillMe->comment[COMMENT_LEN-1] = '\0';
	 strTrimEnd(fillMe->comment);
      }
      else
	 strcpy(fillMe->comment, "");
      /* strcpy(fillMe->currentBoard, goGameBoardString(theGame)); */
      ret = fillMe;
   }

   return ret;
}

int goMessageCompare(goMessage *m1, goMessage *m2)
{
   int compare = m2->player - m1->player;

   if (compare == 0) compare = m2->type - m1->type;
   if (compare == 0) compare = m2->y - m1->y;
   if (compare == 0) compare = m2->x - m1->x;
   return compare;
}

void goMessageSend(goGame *theGame, int reason,
		   int x, int y, int player, char *text)
{
   goMessage fillMe;
   
   goMessageFill(&fillMe, theGame, reason, x, y, player, text);
   goMessageApply(theGame, &fillMe);
}

void recursiveRemoveStone(goGame *theGame, int boardx, int boardy,
			  int removeColor)
{
   int spot;
   
   if (theGame && 
       boardx >= 0 && boardx < theGame->numColumns &&
       boardy >= 0 && boardx < theGame->numRows)
   {
      spot = boardGetSpot(theGame, boardx, boardy);
      if (spot == removeColor)
      {
	 /* its a member of the group */
	 goMessageSend(theGame, MESSAGE_CAPTURE_STONE,
		       boardx, boardy, theGame->me, NULL);
	 /* potentially remove all the neighbor stones */
	 recursiveRemoveStone(theGame, boardx-1, boardy, removeColor);
	 recursiveRemoveStone(theGame, boardx+1, boardy, removeColor);
	 recursiveRemoveStone(theGame, boardx, boardy-1, removeColor);
	 recursiveRemoveStone(theGame, boardx, boardy+1, removeColor);
      }
   }
}

void recursiveTallyLiberties(goGame *theGame, int boardx, int boardy,
			     int groupColor, int *liberties, char *checked)
{
   int spot;

   if (theGame && 
       boardx >= 0 && boardx < theGame->numColumns &&
       boardy >= 0 && boardx < theGame->numRows &&
       checked[(boardy*theGame->numRows)+boardx] == BOARD_EMPTY)
   {
      spot = boardGetSpot(theGame, boardx, boardy);
      if (spot == groupColor)
      {
	 /* its a member of the group */
	 /* add this stones liberties to the sum total */
	 if (boardx-1 >= 0 &&
	     boardGetSpot(theGame, boardx-1, boardy) == PLAYER_NONE)
	    (*liberties)++;
	 if (boardx+1 < theGame->numColumns &&
	     boardGetSpot(theGame, boardx+1, boardy) == PLAYER_NONE)
	    (*liberties)++;
	 if (boardy-1 >= 0 &&
	     boardGetSpot(theGame, boardx, boardy-1) == PLAYER_NONE)
	    (*liberties)++;
	 if (boardy+1 < theGame->numRows &&
	     boardGetSpot(theGame, boardx, boardy+1) == PLAYER_NONE)
	    (*liberties)++;
	 checked[(boardy*theGame->numRows)+boardx] = BOARD_BLACK;
	 /* count liberties of neighbor stones */
	 recursiveTallyLiberties(theGame, boardx-1, boardy,
				 groupColor, liberties, checked);
	 recursiveTallyLiberties(theGame, boardx+1, boardy,
				 groupColor, liberties, checked);
	 recursiveTallyLiberties(theGame, boardx, boardy-1,
				 groupColor, liberties, checked);
	 recursiveTallyLiberties(theGame, boardx, boardy+1,
				 groupColor, liberties, checked);
      }
   }
}

/* return True if a given move forces a capture of a group specified by
 * an x, y stone */

int moveCapturesOpponent(goGame *theGame, goMessage *rascal,
		      int groupx, int groupy)
{
   int ret = False;
   int spot, opponent;
   
   if (theGame && 
       groupx >= 0 && groupx < theGame->numColumns &&
       groupy >= 0 && groupx < theGame->numRows)
   {
      spot = boardGetSpot(theGame, groupx, groupy);
      if (rascal->player == PLAYER_BLACK)
	 opponent = PLAYER_WHITE;
      else if (rascal->player == PLAYER_WHITE)
	 opponent = PLAYER_BLACK;
      if (spot == opponent)
      {
	 int liberties, i;
	 char checked[BOARD_MAX_LEN];

	 /* could be a move that captures a group...  hmmm, let's count the
	  * liberties */
	 liberties = 0;
	 strcpy(checked, theGame->state);
	 for (i = 0; i < strlen(checked); i++) checked[i] = BOARD_EMPTY;
	 recursiveTallyLiberties(theGame, groupx, groupy, spot,
				 &liberties, checked);
	 if (liberties == 0)
	    ret = True;
      }
   }

   return ret;
}

/* returns True if the given move is considered suicide,
 * False if ok...
 */

int moveCapturesSelf(goGame *theGame, goMessage *rascal, int player)
{
   int ret = False;
   int spot;
   
   if (theGame)
   {
      spot = boardGetSpot(theGame, rascal->x, rascal->y);
      if (spot == player)
      {
	 int liberties, i;
	 char checked[BOARD_MAX_LEN];

	 /* could be a move that kills me...  hmmm, let's count the
	  * liberties */
	 liberties = 0;
	 strcpy(checked, theGame->state);
	 for (i = 0; i < strlen(checked); i++) checked[i] = BOARD_EMPTY;
	 recursiveTallyLiberties(theGame, rascal->x, rascal->y,
				 spot, &liberties, checked);
	 if (liberties == 0)
	    ret = True;
      }
   }

   return ret;
}

/* return True if a given move causes a repetition in board positions
 * based on the applyRule:
 * if applyRule == REASON_NORMAL_KO, check for board equality 1 move ago...
 * if applyRule == REASON_SUPER_KO, check for equality 
 */

int boardReproduced(goGame *theGame, goGame *possible, int applyRule)
{
   int ret = False;
   goHistory *lookBack;

#if defined(ILLEGAL_DEBUG)
   printf("possible: (reason %d)\n", applyRule);
   outputBoardText(stdout, possible->state,
		   theGame->numRows, theGame->numColumns);
#endif
   if (theGame &&
       (lookBack = goHistorySearchLast(theGame->historyLast,
				       MESSAGE_PLACE_STONE)))
   {
#if defined(ILLEGAL_DEBUG)
      printf("compare:\n");
      outputBoardText(stdout, lookBack->board,
		      theGame->numRows, theGame->numColumns);
#endif
      if (applyRule == REASON_NORMAL_KO)
	 ret = (strcmp(lookBack->board, possible->state)==0? True: False);
      else if (applyRule == REASON_SUPER_KO)
      {
	 while (lookBack && !ret)
	 {
	    ret = (strcmp(lookBack->board, possible->state)==0? True: False);
	    lookBack = goHistorySearchLast(lookBack->prev,
					   MESSAGE_PLACE_STONE);
	 }
      }
   }

   return ret;
}

char *goMessageReasonStr(int reason)
{
   char *errmsg = "Unknown reason";

   switch (reason)
   {
    case REASON_VALID: /* shouldn't happen */
      break;
    case REASON_UNKNOWN:
      errmsg = "Move rejected for unknown reason, sorry...";
      break;
    case REASON_OBSERVER:
      errmsg = "You are only an observer.";
      break;
    case REASON_NO_PLAYER:
      errmsg = "You haven't selected your stone color yet.";
      break;
    case REASON_SUICIDE:
      errmsg = "That move is an illegal suicide.";
      break;
    case REASON_NORMAL_KO:
      errmsg = "That move is an illegal ko capture.";
      break;
    case REASON_SUPER_KO:
      errmsg = "That move is an illegal (super) ko capture.";
      break;
    case REASON_GAME_OVER:
      errmsg = "Both players have passed.  Undo before moving again.";
      break;
   }
   return errmsg;
}


/* returns True if the message is valid according to the go rules
 * returns False if the move is not allowed
 * reason is the reason the move is invalid
 * note: only place_stone types of messages need validation
 */

int goMessageValid(goGame *theGame, goMessage *message, int *reason)
{
   int ret = True;
   int retReason = REASON_VALID;

   if (theGame && message && message->type == MESSAGE_PLACE_STONE
       && !(message->override))
   {
      goGame *copy = goGameCopy(theGame, False);
      int player = theGame->me;

      goMessageApply(copy, message);
      /* placements of stones are the only questionable moves... */
      ret = False;
      if (player == PLAYER_NONE)
	 retReason = REASON_NO_PLAYER;
      else if (player == PLAYER_OBSERVER)
	 retReason = REASON_OBSERVER;
      else if (moveCapturesSelf(copy, message, player))
	 retReason = REASON_SUICIDE;
      else if (!goGameGetOption(theGame, OPTION_SUPER_KO_OFF) &&
	       boardReproduced(theGame, copy, REASON_SUPER_KO))
	 retReason = REASON_SUPER_KO;
      else if (boardReproduced(theGame, copy, REASON_NORMAL_KO))
	 retReason = REASON_NORMAL_KO;
      else if (goGameGetNumPasses(theGame) >= 2)
	 retReason = REASON_GAME_OVER;
      else
	 ret = True;

      goGameDestroy(copy);
   }
   
   if (reason) *reason = retReason;
   return ret;
}

int undoLastMove(goGame *theGame, int *opponent)
{
   goHistory *hist = NULL;
   int ret = False, me;
   
   *opponent = PLAYER_NONE;
   if (theGame && theGame->keepHistory)
   {
      hist = theGame->historyLast;
      me = theGame->me;
   }

   goGameSetKeepHistory(theGame, False);
   while (hist && hist->turn &&
	  hist->turn->type != MESSAGE_PLACE_STONE &&
	  hist->turn->type != MESSAGE_CAPTURE_GROUP)
   {
      int isCapture = (hist->turn->type == MESSAGE_CAPTURE_STONE);
      int isPass = (hist->turn->type == MESSAGE_PASS);
      goMessage *mess = goMessageCopy(hist->turn);
      goHistory *prev = hist->prev;
      
      if (mess)
      {
	 /* turn history off -- don't record the place stones */
	 goGameRemoveFromHistory(theGame, hist->turn);
	 if (isCapture)
	 {
	    /* re-place all those stones */
	    mess->type = MESSAGE_PLACE_STONE;
	    /* don't let this message capture surrounding groups */
	    mess->override = (char)True;	
	    goMessageApply(theGame, mess);
	 }
	 else if (isPass) theGame->numPasses--;
	 goMessageDestroy(mess);
      }
      hist = prev;
   }
   if (hist && hist->turn &&
       (hist->turn->type == MESSAGE_PLACE_STONE || 
	hist->turn->type == MESSAGE_CAPTURE_GROUP))
   {
      int wasPlace = (hist->turn->type == MESSAGE_PLACE_STONE);
      goMessage *mess = goMessageCopy(hist->turn);

      if (mess)
      {
	 goGameRemoveFromHistory(theGame, hist->turn);
	 if (wasPlace)
	 {
	    /* remove the last placed stone */
	    mess->type = MESSAGE_CAPTURE_STONE;
	    goMessageApply(theGame, mess);
	 }
	 *opponent = mess->player;
	 goMessageDestroy(mess);
	 ret = True;
      }
   }
   goGameSetKeepHistory(theGame, True);
   return ret;
}

/* returns True if the message is valid according to the go rules
 * returns False if error or the move is not allowed
 */

void goMessageApply(goGame *theGame, goMessage *message)
{
   if (theGame && message)
   {
      goMessage *copy = NULL;
      int opponent = PLAYER_NONE;

      /* only keep track of history if user desires */
      if (theGame->keepHistory)
      {
	 switch (message->type)
	 {
	  case MESSAGE_COMMENT:
	  case MESSAGE_PLACE_STONE:
	  case MESSAGE_CAPTURE_STONE:
	  case MESSAGE_PASS:
	    copy = goMessageCopy(message);
	    goGameAddToHistory(theGame, copy);
	    break;
	  case MESSAGE_CAPTURE_GROUP:
	    /* this is a special case... if the game is over record the
	     * message */
	    if (goGameGetNumPasses(theGame) >= 2)
	    {
	       copy = goMessageCopy(message);
	       goGameAddToHistory(theGame, copy);
	    }
	    /* otherwise, don't */
	    break;
	  default:
	    /* no need to keep history of other these events... */
	    break;
	 }
      }

      /* internally update the status of the game --
       * assume the move/capture is valid... */
      if (message->player == PLAYER_BLACK)
	 opponent = PLAYER_WHITE;
      else if (message->player == PLAYER_WHITE)
	 opponent = PLAYER_BLACK;
      switch (message->type)
      {
       case MESSAGE_PLACE_STONE:
	 boardSetSpot(theGame, message->x, message->y, message->player);
	 /* only see if this captures a group on non-overrides... */
	 if (!message->override)
	 {
	    if (moveCapturesOpponent(theGame, message,
				     message->x-1, message->y))
	       goMessageSend(theGame, MESSAGE_CAPTURE_GROUP,
			     message->x-1, message->y, opponent, NULL);
	    if (moveCapturesOpponent(theGame, message,
				     message->x+1, message->y))
	       goMessageSend(theGame, MESSAGE_CAPTURE_GROUP,
			     message->x+1, message->y, opponent, NULL);
	    if (moveCapturesOpponent(theGame, message,
				     message->x, message->y-1))
	       goMessageSend(theGame, MESSAGE_CAPTURE_GROUP,
			     message->x, message->y-1, opponent, NULL);
	    if (moveCapturesOpponent(theGame, message,
				     message->x, message->y+1))
	       goMessageSend(theGame, MESSAGE_CAPTURE_GROUP,
			     message->x, message->y+1, opponent, NULL);
	 }
	 /* else printf("%s\n", goMessageStr(message)); */
	 goGameSetTurn(theGame, opponent);
	 break;
       case MESSAGE_PASS:
	 theGame->numPasses++;
	 goGameSetTurn(theGame, opponent);
	 break;
       case MESSAGE_CAPTURE_STONE:
	 /* record the color of the captured stone (leave owner alone) */
	 /* this is kind of wierd, modifying the copy information
	  * after the fact (its already been added to history) */
	 if (copy)
	    copy->player = boardGetSpot(theGame, message->x, message->y);
	 boardSetSpot(theGame, message->x, message->y, PLAYER_NONE);
	 break;
       case MESSAGE_CAPTURE_GROUP:
	 if (!message->override)
	    recursiveRemoveStone(theGame, message->x, message->y,
				 message->player);
	 break;
       case MESSAGE_UNDO_MOVE:
	 undoLastMove(theGame, &opponent);
	 if (opponent != PLAYER_NONE)
	    goGameSetTurn(theGame, opponent);
	 break;
       case MESSAGE_NEW_GAME:
	 goGameClear(theGame);
	 goGameSetTurn(theGame, PLAYER_BLACK);
	 break;
       default:
	 /* other messages have no effect on the game */
	 break;
      }
   }
}

