/* goScore.c -- uses stone radiation of influence to calculate score */

/* mango/tango -- copyright 1994 Edward Grigassy
 * the code included in this package may be distributed, read, and copied
 * freely BUT NOT SOLD OR INCLUDED IN OTHER SOFTWARE, for commercial or
 * public/free use, without express written consent of the author.
 */

#include "public.h"
#include "strmisc.h"

#include "goGameP.h"
#include "goScore.h"

/* ------------------------------------------------------------------------- */
/* typedefs, struct definitions */
/* ------------------------------------------------------------------------- */

struct _goScore {
   goGame *game;
   int numRows, numColumns;
   char playerBlack, playerWhite, emptySpace;
   char thisPlayer;
   float value[BOARD_MAX_LEN];
   char influenced[BOARD_MAX_LEN];
   char board[BOARD_MAX_LEN];
   double total;
};

struct _goBoardXY { char x, y; };

struct _goTerritory {
   goGame *game;
   int numRows, numColumns;
   int player, opponent;
   struct _goBoardXY mine[BOARD_MAX_LEN];
   struct _goBoardXY yours[BOARD_MAX_LEN];
   struct _goBoardXY neutral[BOARD_MAX_LEN];
   int numMineEmpty, numYoursEmpty, numNeutral;
   int numMineCaptured, numYoursCaptured;
   int numMinePlaced, numYoursPlaced;
   int myTotal, yourTotal;
};

/* ------------------------------------------------------------------------- */
/* globals */
/* ------------------------------------------------------------------------- */

int decayMultiple = 1, influenceDistance = 4, ownershipMult = 2;
float ownershipFloor = 0.0;

int penaltiesInitialized = False;
int boardNumRows = 0, boardNumColumns = 0;
float distancePenalties[BOARD_MAX_LEN];
float emptyBoardValues[BOARD_MAX_LEN];

/* ------------------------------------------------------------------------- */
/* goTerritory routines */
/* ------------------------------------------------------------------------- */

goTerritory *goTerritoryCreate(goGame *game, int player)
{
   goTerritory *ret = NULL;

   if (ret = NEW(goTerritory))
   {
      int size = game->numColumns*game->numRows;
      int clean;
      
      ret->game = game;
      ret->numColumns = game->numColumns;
      ret->numRows = game->numRows;
      ret->player = (player != PLAYER_BLACK && player != PLAYER_WHITE?
		     PLAYER_BLACK: player);
      ret->opponent = (ret->player == PLAYER_BLACK?
		     PLAYER_WHITE: PLAYER_BLACK);
   }
   return ret;
}

goTerritory *goTerritoryDestroy(goTerritory *territory)
{
   if (territory)
   {
      DESTROY(territory);
   }
   return NULL;
}

void goTerritorySortSpot(goTerritory *terr, int i, int j, int owner)
{
   if (owner == terr->player)
   {
      terr->mine[terr->numMineEmpty].x = (char)i;
      terr->mine[terr->numMineEmpty].y = (char)j;
      terr->numMineEmpty++;
   }
   else if (owner == terr->opponent)
   {
      terr->yours[terr->numYoursEmpty].x = (char)i;
      terr->yours[terr->numYoursEmpty].y = (char)j;
      terr->numYoursEmpty++;
   }
   else
   {
      terr->neutral[terr->numNeutral].x = (char)i;
      terr->neutral[terr->numNeutral].y = (char)j;
      terr->numNeutral++;
   }
}
		     
void goTerritoryInitialize(goTerritory *terr)
{
   terr->numMineEmpty = 0;
   terr->numYoursEmpty = 0;
   terr->numNeutral = 0;
   terr->numMineCaptured = 0;
   terr->numYoursCaptured = 0;
   terr->numMinePlaced = 0;
   terr->numYoursPlaced = 0;
   terr->myTotal = 0;
   terr->yourTotal = 0;
}

int getNonEmptySpot(goGame *game, int i, int j, int imod, int jmod)
{
   int ret = PLAYER_NONE;

   while (ret == PLAYER_NONE && i >= 0 && i < game->numColumns &&
	  j >= 0 && j < game->numRows)
   {
      ret = boardGetSpot(game, i, j);
      i += imod;
      j += jmod;
   }
   return ret;
}

/* determine who owns this spot */
/* simple algorithm: surrounded in all four directions by one player */
/* in the four directions there can be as many as 3 no-players,
 * but to be "owned" a spot must have at least one player in a direction
 * and total of only one player */

int goTerritoryGetOwner(goGame *game, int i, int j)
{
   int card[4], loop, livePlayer = PLAYER_NONE;
   int ret = PLAYER_NONE;

   card[0] = getNonEmptySpot(game, i, j-1, 0, -1);	/* up */
   card[1] = getNonEmptySpot(game, i, j+1, 0, 1);	/* down */
   card[2] = getNonEmptySpot(game, i-1, j, -1, 0);	/* left */
   card[3] = getNonEmptySpot(game, i+1, j, 1, 0);	/* right */
   for (loop = 0; loop < 4; loop++)
   {
      if (card[loop] != PLAYER_NONE) livePlayer = card[loop];
   }
   if (livePlayer != PLAYER_NONE)
   {
      int allSame = True;
      
      for (loop = 0; loop < 4; loop++)
      {
	 if (card[loop] != PLAYER_NONE && card[loop] != livePlayer)
	    allSame = False;
      }
      if (allSame)
	 ret = livePlayer;
   }
   return ret;
}

int goTerritoryCalculate(goTerritory *terr, int ruleType)
{
   int winner = PLAYER_NONE;

   if (terr && terr->game)
   {
      int i, j;

      goTerritoryInitialize(terr);
      
      /* determine territory of all empty spots */
      for (j = 0; j < terr->numRows; j++)
      {
	 for (i = 0; i < terr->numColumns; i++)
	 {
	    int spot = boardGetSpot(terr->game, i, j);
	    int owner;

	    if (spot == PLAYER_NONE)
	    {
	       owner = goTerritoryGetOwner(terr->game, i, j);
	       goTerritorySortSpot(terr, i, j, owner);
	    }
	 }
      }

      /* determine number of stones placed / captured for each player */
      if (terr->game->keepHistory)
      {
	 goHistory *parse;
	 
	 for (parse = terr->game->historyFirst; parse; parse = parse->next)
	 {
	    int who = parse->turn->player;

	    if (parse->turn->type == MESSAGE_PLACE_STONE)
	    {
	       if (who == terr->player) terr->numMinePlaced++;
	       else if (who == terr->opponent) terr->numYoursPlaced++;
	    }
	    if (parse->turn->type == MESSAGE_CAPTURE_STONE)
	    {
	       if (who == terr->player) terr->numMineCaptured++;
	       else if (who == terr->opponent) terr->numYoursCaptured++;
	    }
	 }
      }

      /* calcuate totals and winner */
      if (ruleType == RULE_TYPE_JAPAN)
      {
	 terr->myTotal = terr->numMineEmpty - terr->numMineCaptured;
	 terr->yourTotal = terr->numYoursEmpty - terr->numYoursCaptured;
      }
      else if (ruleType == RULE_TYPE_CHINA)
      {
	 /* for chinese rules its the number of stones you have left on
	  * the board + the number of empty territory you surround */
	 terr->myTotal = terr->numMinePlaced - terr->numMineCaptured
	    + terr->numMineEmpty;
	 terr->yourTotal = terr->numYoursPlaced - terr->numYoursCaptured
	    + terr->numYoursEmpty;
      }
      if (terr->myTotal > terr->yourTotal) winner = terr->player;
      else if (terr->myTotal < terr->yourTotal) winner = terr->opponent;
   }
   
   return winner;
}

int goTerritoryPrintTerritoryList(goTerritory *terr)
{
   if (terr)
   {
      char *p = playerNames[terr->player], *o = playerNames[terr->opponent];
      int i;

      printf("%s territory:\n", p);
      for (i = 0; i < terr->numMineEmpty; i++)
	 printf("(%d,%d) ", terr->mine[i].x, terr->mine[i].y);
      printf("\n");
      printf("%s territory:\n", o);
      for (i = 0; i < terr->numYoursEmpty; i++)
	 printf("(%d,%d) ", terr->yours[i].x, terr->yours[i].y);
      printf("\n");
   }
}
int goTerritoryPrint(goTerritory *terr)
{
   if (terr)
   {
      char *p = playerNames[terr->player], *o = playerNames[terr->opponent];
      int i;
      
      printf("neutral territory: %d\n", terr->numNeutral);
      printf("%s: territory: %d, captured: %d, placed: %d\n", p,
	     terr->numMineEmpty, terr->numMineCaptured, terr->numMinePlaced);
      printf("%s: total score: %d\n", p, terr->myTotal);
      printf("%s: territory: %d, captured: %d, placed: %d\n", o,
	     terr->numYoursEmpty, terr->numYoursCaptured,
	     terr->numYoursPlaced);
      printf("%s: total score: %d\n", o, terr->yourTotal);
   }
}

int goTerritoryGetPlaced(goTerritory *terr, int player)
{
   int ret = 0;
   if (terr)
   {
      if (player == terr->player) ret = terr->numMinePlaced;
      if (player == terr->opponent) ret = terr->numYoursPlaced;
   }
   return ret;
}

int goTerritoryGetCaptured(goTerritory *terr, int player)
{
   int ret = 0;
   if (terr)
   {
      if (player == terr->player) ret = terr->numMineCaptured;
      if (player == terr->opponent) ret = terr->numYoursCaptured;
   }
   return ret;
}

int goTerritoryGetSurrounded(goTerritory *terr, int player)
{
   int ret = 0;
   if (terr)
   {
      if (player == terr->player) ret = terr->numMineEmpty;
      if (player == terr->opponent) ret = terr->numYoursEmpty;
   }
   return ret;
}

int goTerritoryGetNeutral(goTerritory *terr)
{
   int ret = 0;
   if (terr)
   {
      ret = terr->numNeutral;
   }
   return ret;
}

int goTerritoryGetTotal(goTerritory *terr, int player)
{
   int ret = 0;
   if (terr)
   {
      if (player == terr->player) ret = terr->myTotal;
      if (player == terr->opponent) ret = terr->yourTotal;
   }
   return ret;
}

/* ------------------------------------------------------------------------- */
/* goScore routines */
/* ------------------------------------------------------------------------- */

float influenceAmount(int distance)
{
   int dividend = distance * decayMultiple;

   if (dividend == 0) dividend = 1;
   return (1.0 / dividend);
}

void influenceEmptyBoardSpot(int x1, int y1, int x2, int y2)
{
   int distance = abs(x1-x2) + abs(y1-y2);
   int spot = y1 * boardNumRows + x1;
	    
   if (distance < influenceDistance)
      emptyBoardValues[spot] += distancePenalties[distance];
}

void initializeBoardScore()
{
   int distance;
   int i, x1, y1, x2, y2, j;
   int spot, boardSize = boardNumRows * boardNumColumns;
   int times = ownershipMult;

   for (distance = 0; distance < BOARD_MAX_LEN; distance++)
   {
      distancePenalties[distance] = influenceAmount(distance);
      if (distancePenalties[distance] > ownershipFloor)
      {
	 for (j = 0; j < times; j++)
	    distancePenalties[distance] *= distancePenalties[distance];
      }
      printf("dist, penalty: %d, %f\n", distance,
	     distancePenalties[distance]);
   }
   
   for (i = 0; i < boardSize; i++)
   {
      emptyBoardValues[i] = 0.0;
   }

   for (y1 = 0; y1 < boardNumRows; y1++)
   {
      for (x1 = 0; x1 < boardNumColumns; x1++)
      {
	 /* add influence for edges (as if there were a ring of the
	  * player's stones around the edges of the board)
	  */

	 for (y2 = 0; y2 < boardNumRows; y2++)
	 {
	    influenceEmptyBoardSpot(x1, y1, -1, y2);
	    influenceEmptyBoardSpot(x1, y1, boardNumRows, y2);
	 }
	 for (x2 = 0; x2 < boardNumColumns; x2++)
	 {
	    influenceEmptyBoardSpot(x1, y1, x2, -1);
	    influenceEmptyBoardSpot(x1, y1, x2, boardNumColumns);
	 }
      }
   }
}

/* algorithm for casting influence:
 * for each stone of given color:
 * distance = 0;
 * start with the stone:
 *   set influence on that spot += influenceAmount(distance);
 * distance++;
 * in 8 directions (4 cardinal directions + 4 diagonal directions)
 * spot += influenceAmount(distance);
 * if spot is not my player or his player, (an empty spot)
 *    recursively determine influence with current spot, distance
 * note: this algorithm emulates the fact that a stone eminates
 * influence up to and including any other stone.  thus if the influence
 * is stopped by a stone of the same color, we are guaranteed that
 * that stones influence will continue the influence of the color;
 * and if the influence is stopped by an opponent's stone, his
 * influence will be continued.
 */

/* new improved algorithm for casting influence:
 * given color COLOR, for each spot on the board
 * go through every other spot on the board that is == COLOR
 * distance = abs(y1-y2)+abs(x1-x2)
 * spot += influenceAmount(distance)
 */

/* given an allocated score structure, revalue it for a given player
 * calculates edges first (top to bottom, left to right) then
 * for each stone the player has
 */

void goScoreCalculateInfluence(goScore *score)
{
   if (score)
   {
      register int x1, y1;
      int x2, y2, spot;
      char atSpot;
      int distance, spotOrig;

      bcopy(emptyBoardValues, score->value, sizeof(emptyBoardValues));

      /* add influence for each stone the player has */
      for (y2 = 0; y2 < score->numRows; y2++)
      {
	 for (x2 = 0; x2 < score->numColumns; x2++)
	 {
	    spot = y2 * score->numRows + x2;
	    atSpot = score->board[spot];
	    if (atSpot == score->thisPlayer)
	    {
	       /* add this influence to every position on the board */
	       for (y1 = 0; y1 < score->numRows; y1++)
	       {
		  for (x1 = 0; x1 < score->numColumns; x1++)
		  {
		     distance = abs(x1-x2) + abs(y1-y2);
		     spotOrig = y1*score->numRows + x1;
		     
		     if (distance < influenceDistance)
			score->value[spotOrig] += distancePenalties[distance];
		  }
	       }
	    }
	 }
      }
   }
}

void goScoreCombine(goScore *comb, goScore *neg, goScore *pos)
{
   int sizeneg = neg->numRows * neg->numColumns;
   int sizepos = pos->numRows * pos->numColumns;
   if (sizeneg == sizepos)
   {
      int i, j;
      int times = ownershipMult;

      neg->total = 0.0;
      pos->total = 0.0;
      comb->total = 0.0;
      for (i = 0; i < sizeneg; i++)
      {
#if defined(NOT_MULT_IN_EMPTY)
	 if (pos->value[i] > ownershipFloor)
	 {
	    for (j = 0; j < times; j++)
	       pos->value[i] *= pos->value[i];
	 }
	 if (neg->value[i] > ownershipFloor)
	 {
	    for (j = 0; j < times; j++)
	       neg->value[i] *= neg->value[i];
	 }
#endif
	 comb->value[i] = pos->value[i] - neg->value[i];
	 /* only total influences of other stones and emtpy space */
	 if (neg->board[i] != neg->thisPlayer)
	    neg->total += neg->value[i];
	 if (pos->board[i] != pos->thisPlayer)
	    pos->total += pos->value[i];
      }
      comb->total = pos->total - neg->total;
   }
}

void goScorePrintCombined(goScore *score, char neg, char pos)
{
   int x, y;
   
   printf("Combined score: (total %lf)\n", score->total);
   for (y = 0; y < score->numRows; y++)
   {
      for (x = 0; x < score->numColumns; x++)
      {
	 int infl = (int)score->value[y*score->numRows+x];
	 int i;
	 char inflStr[6], posch = pos, negch = neg;
	 
	 if (infl < -4) infl = -4, negch = toupper(neg);
	 if (infl > 4) infl = 4, posch = toupper(pos);
	 strcpy(inflStr, "-   ");
	 for (i = 0; i < abs(infl); i++)
	 {
	    inflStr[i] = (infl > 0? pos: neg);
	 }
	 printf("%s ", inflStr);
      }
      printf("\n");
   }
}

void goScorePrint(goScore *score)
{
   int x, y;
   
   printf("%c to move.\n", score->thisPlayer);
   for (y = 0; y < score->numRows; y++)
   {
      for (x = 0; x < score->numRows; x++)
      {
	 printf("%1.2lf ", score->value[y*score->numRows+x]);
      }
      printf("\n");
   }
}

goScore *goScoreDestroy(goScore *score)
{
   if (score)
   {
      DESTROY(score);
   }
   return NULL;
}

goScore *goScoreCreate(goGame *game, int player)
{
   goScore *ret = NULL;

   if (!penaltiesInitialized)
   {
      boardNumRows = game->numRows;
      boardNumColumns = game->numColumns;
      initializeBoardScore();
      penaltiesInitialized = True;
   }
   if (ret = NEW(goScore))
   {
      int size = game->numColumns*game->numRows;
      int clean;
      
      ret->game = game;
      ret->numColumns = game->numColumns;
      ret->numRows = game->numRows;
      ret->playerBlack = BOARD_BLACK;
      ret->playerWhite = BOARD_WHITE;
      ret->emptySpace = BOARD_EMPTY;
      strcpy(ret->board, game->state);
      for (clean = 0; clean < size; clean++)
      {
	 ret->value[clean] = 0.0;
	 ret->influenced[clean] = False;
      }
      if (player == PLAYER_WHITE)
	 ret->thisPlayer = BOARD_WHITE;
      else if (player == PLAYER_BLACK)
	 ret->thisPlayer = BOARD_BLACK;
      else
	 ret->thisPlayer = BOARD_EMPTY;
   }
   return ret;
}

double goScoreGetValue(goScore *score)
{
   double ret = -1.0;

   if (score)
      ret = score->total;
   return ret;
}

