/* xIFace.c -- handles the go interface of choice (X windows) *

/* xmango -- copyright 1994 Edward Grigassy
 * the code included in this package may be distributed, read, and copied
 * freely BUT NOT SOLD OR INCLUDED IN OTHER SOFTWARE, for commercial or
 * public/free use, without express written consent of the author.
 */

/* include files: */
#include "xgo.h"

#include "goGameP.h"
#include "xIFaceP.h"
#include "goScore.h"
#include "dialogAPI.h"

/* ------------------------------------------------------------------------- */
/* goGameInterface API routines */
/* ------------------------------------------------------------------------- */

goGameInterface *goGameInterfaceCreate(int rows, int columns,
				       int starPoint, int turn,
				       int rules)
{
   goGameInterface *ret = NULL;
   goInterface *iface = NULL;
   goGame *game = NULL;

   iface = NEW(goInterface);
   game = goGameCreate(rows, columns, starPoint, turn);
   if (iface && game)
   {
      ret = NEW(goGameInterface);
      ret->iface = iface;
      ret->game = game;
      ret->scoringRules = rules;
      ret->suppressOutput = False;
   }
   
   return ret;
}

goGameInterface *goGameInterfaceDestroy(goGameInterface *ggi)
{
   if (ggi)
   {
      if (ggi->iface->removeCursor)
	 /* XFreeCursor(); */
      DESTROY(ggi->iface);
      ggi->game = goGameDestroy(ggi->game);
   }
   
   return NULL;
}

Widget goGameInterfaceGetBoard(goGameInterface *ggi)
{
   Widget ret = NULL;
   
   if (ggi && ggi->iface)
      ret = ggi->iface->area;
   return ret;
}

int goGameInterfaceSetOption(goGameInterface *ggi, long int option, int high)
{
   return goGameSetOption(ggi->game, option, high);
}

int goGameInterfaceGetOption(goGameInterface *ggi, long int option)
{
   return goGameGetOption(ggi->game, option);
}

int goGameInterfaceSetId(goGameInterface *ggi, int myId)
{
   return goGameSetId(ggi->game, myId);
}

int goGameInterfaceDisconnect(goGameInterface *ggi)
{
   int ret = False;
   if (ggi && goGameInterfaceConnectionType(ggi) == CONNECTION_SOCKET)
   {
      if (ggi->socketInput)
      {
	 XtRemoveInput(ggi->socketInput);
	 ggi->socketInput = 0;
      }
      closeSocket(goGameInterfaceGetConnection(ggi));
      goGameInterfaceSetConnection(ggi, NULL);
      goGameInterfaceSetConnectionType(ggi, CONNECTION_SINGLE_INSTANCE);
      ret = True;
   }
   return ret;
}

void goInterfaceSetColors(goGameInterface *ggi,
			  unsigned long grid,
			  unsigned long shadow,
			  unsigned long bright,
			  unsigned long light,
			  unsigned long dull,
			  unsigned long player1,
			  unsigned long player2)
{
   ggi->iface->grid = grid;
   ggi->iface->shadow = shadow;
   ggi->iface->bright = bright;
   ggi->iface->light = light;
   ggi->iface->dull = dull;
   ggi->iface->player1 = player1;
   ggi->iface->player2 = player2;
}

int goGameInterfaceConnectionType(goGameInterface *ggi)
{
   int ret = CONNECTION_SINGLE_INSTANCE;
   if (ggi)
      ret = ggi->connectionType;
   return ret;
}

int goGameInterfaceSetConnectionType(goGameInterface *ggi, int connectionType)
{
   int ret = -1;
   if (ggi) ret = ggi->connectionType = connectionType;
   return ret;
}

qHead *goGameInterfaceGetConnection(goGameInterface *ggi)
{
   qHead *ret = NULL;
   if (ggi) ret = ggi->connection;
   return ret;
}

int goGameInterfaceSetConnection(goGameInterface *ggi, qHead *connection)
{
   int ret = -1;
   if (ggi)
   {
      ggi->connection = connection;
      ret = 0;
   }
   return ret;
}

/* ------------------------------------------------------------------------- */
/* goGameInterface event routines */
/* ------------------------------------------------------------------------- */

void setRemoveButtonSensitive(goGameInterface *ggi)
{
   if (goGameGetNumPasses(ggi->game) >= 2)
      XtSetSensitive(ggi->iface->removeButton, True);
   else
      XtSetSensitive(ggi->iface->removeButton, False);
}

void showTurnLabels(goGameInterface *ggi)
{
   if (goGameInterfaceConnectionType(ggi) == CONNECTION_SINGLE_INSTANCE)
   {
      goGameSetPlayer(ggi->game, goGameGetTurn(ggi->game));
      setMeLabel(ggi);
   }
   setTurnLabel(ggi);
}

void showInvalidReason(goGameInterface *ggi, int reason)
{
   char *errmsg = goMessageReasonStr(reason);
   
   if (ggi && ggi->iface && errmsg)
   {
      dialogErrorPopup(ggi->iface->area,"Move Error",errmsg,False);
      beepGame(ggi);
   }
}

void undoMoveCB(Widget w, goGameInterface *ggi, XtPointer nothing)
{
   if (ggi)
   {
      /* for socket connection games only allow undo by the
       * user who made the last placement */
      if (goGameInterfaceConnectionType(ggi) == CONNECTION_SINGLE_INSTANCE)
      {
	 ggiMessageSend(ggi, True, MESSAGE_UNDO_MOVE,
			0, 0, goGameGetPlayer(ggi->game), NULL);
      }
      else if (goGameGetTurn(ggi->game) != goGameGetPlayer(ggi->game))
	 ggiMessageSend(ggi, True, MESSAGE_UNDO_MOVE,
			0, 0, goGameGetPlayer(ggi->game), NULL);
      else
	 ggiMessageSend(ggi, True, MESSAGE_REQUEST_UNDO,
			0, 0, goGameGetPlayer(ggi->game), NULL);
   }
}

/* ------------------------------------------------------------------------- */
/* territory statistics display routines */
/* ------------------------------------------------------------------------- */

void territoryOff(Widget w, goGameInterface *ggi, XtPointer whatYouLookinAt)
{
   ggiMessageSend(ggi, False, MESSAGE_TERRITORY_OFF,
		  0, 0, PLAYER_NONE, NULL);
}

Widget territoryPanelBuild(goGameInterface *ggi)
{
   Widget shell = widget_to_shell(goGameInterfaceGetBoard(ggi));
   char *title = "Current Score", scoreInfo[512];
   dialogType *dialog = NULL;

   if (ggi && ggi->territory)
   {
      char bname[20], wname[20], *ganarian, *ruleString, *winnerString;
      int bp, bc, bs, bt, wp, wc, ws, wt, ns, winner;
      
      /* fill in score information */
      ganarian = (goGameGetNumPasses(ggi->game) >= 2? "Winner": "Leader");
      ruleString = (ggi->scoringRules == RULE_TYPE_CHINA? "China": "Japan");
      strcpy(bname, playerNames[PLAYER_BLACK]);
      bname[0] = toupper(bname[0]);
      strcpy(wname, playerNames[PLAYER_WHITE]);
      wname[0] = toupper(wname[0]);
      winner = goTerritoryCalculate(ggi->territory, ggi->scoringRules);
      bp = goTerritoryGetPlaced(ggi->territory, PLAYER_BLACK);
      bc = goTerritoryGetCaptured(ggi->territory, PLAYER_BLACK);
      bs = goTerritoryGetSurrounded(ggi->territory, PLAYER_BLACK);
      bt = goTerritoryGetTotal(ggi->territory, PLAYER_BLACK);
      wp = goTerritoryGetPlaced(ggi->territory, PLAYER_WHITE);
      wc = goTerritoryGetCaptured(ggi->territory, PLAYER_WHITE);
      ws = goTerritoryGetSurrounded(ggi->territory, PLAYER_WHITE);
      wt = goTerritoryGetTotal(ggi->territory, PLAYER_WHITE);
      ns = goTerritoryGetNeutral(ggi->territory);
      winnerString = (winner == PLAYER_WHITE? wname: bname);
      sprintf(scoreInfo, "Scoring Rules: %s\n \n"
	      "Stones/Intersection Information:\n"
	      "Player   Played  Captured  Surrounded  Score\n"
	      "%-7s  %6d  %8d  %10d  %5d\n"
	      "%-7s  %6d  %8d  %10d  %5d\n"
	      "Neutral                    %10d       \n \n"
	      "%s: %s",
	      ruleString,
	      bname, bp, bc, bs, bt,
	      wname, wp, wc, ws, wt,
	      ns,
	      ganarian, winnerString);
      dialog = dialogCreate(shell, DIALOG_INFORMATION, title, True);
   }
   if (dialog)
   {
      dialogSetLabel(dialog, scoreInfo);
      /*
      dialogAddButton(dialog, BUTTON_OK, "Black",
		      NULL, ggi, False);
      dialogAddButton(dialog, BUTTON_APPLY, "White",
		      NULL, ggi, False);
      dialogAddButton(dialog, BUTTON_CANCEL, "Neutral",
		      NULL, ggi, False);
		      */
      dialogAddButton(dialog, BUTTON_HELP, "Close",
		      (XtCallbackProc)territoryOff, ggi, True);
      dialogPopup(dialog);
      
   }
}

/* ------------------------------------------------------------------------- */
/* ggiMessage handling routines */
/* ------------------------------------------------------------------------- */

void ggiMessageSend(goGameInterface *ggi, int maybeBeep, int reason,
		    int x, int y, int player, char *text)
{
   goMessage fillMe;
   
   goMessageFill(&fillMe, ggi->game, reason, x, y, player, text);
   ggiMessageApply(ggi, &fillMe, maybeBeep);
}


void ggiMessageApply(goGameInterface *ggi, goMessage *message,
		     int beepOverride)
{
   char messtr[128], playerName[40];
   goGame *copy = goGameCopy(ggi->game, False);
   goMessage *transfer;

   /* send same message to opponent if I generated it */
   if (goGameInterfaceConnectionType(ggi) == CONNECTION_SOCKET &&
       message->owner == goGameGetId(ggi->game))
   {
      /* printf("sending: %s\n", goMessageStr(message)); */
      sendSocket(ggi->connection, message, sizeof(goMessage));
   }

   /* apply the message to the game first... */
   goMessageApply(ggi->game, message);

   if (message->comment)
   {
      if (message->owner == goGameGetId(ggi->game))
	 sprintf(messtr,"%s\n", message->comment);
      else
	 sprintf(messtr,"> %s\n", message->comment);
   }

   /* sync the sender of the transfer */
   if (ggi->suppressOutput)
   {
      goMessage temp;

      goMessageFill(&temp, ggi->game, MESSAGE_REQUEST_MOVE, 0, 0,
		    goGameGetPlayer(ggi->game), NULL);
      sendSocket(ggi->connection, &temp, sizeof(goMessage));
   }
      
   switch (message->type)
   {
    case MESSAGE_PLACE_STONE:
      /* assume that the placement is valid */
      if (!ggi->suppressOutput)
      {
	 drawBoardDifferent(ggi, copy);
	 if (beepOverride &&
	     goGameInterfaceGetOption(ggi, OPTION_BEEP_ON_MOVE))
	    beepGame(ggi);
	 highlightLastStone(ggi);
	 showTurnLabels(ggi);
	 if (goGameInterfaceConnectionType(ggi) == CONNECTION_SOCKET &&
	     message->owner == goGameGetId(ggi->game))
	 {
	    ggiMessageSend(ggi, False, MESSAGE_REQUEST_MOVE,
			   message->x, message->y, goGameGetPlayer(ggi->game),
			   NULL);
	 }
      }
      break;
    case MESSAGE_REQUEST_MOVE:
      break;
    case MESSAGE_PASS:
      /* some messages that come across the socket may be suppressed */
      if (!ggi->suppressOutput)
      {      
	 setRemoveButtonSensitive(ggi);
	 showTurnLabels(ggi);
      }
      strcpy(playerName, playerNames[message->player]);
      playerName[0] = toupper(playerName[0]);
      sprintf(messtr,"* %s passes.\n", playerName);
      commentAreaAppend(ggi, messtr);
      break;
    case MESSAGE_CAPTURE_STONE:
      if (!ggi->suppressOutput)
      {
	 drawIntersection(ggi, message->x, message->y);
      }
      break;
    case MESSAGE_CAPTURE_GROUP:
      if (!ggi->suppressOutput)
      {
	 drawBoardDifferent(ggi, copy);
      }
      break;
    case MESSAGE_UNDO_MOVE:
      drawBoardDifferent(ggi, copy);
      highlightLastStone(ggi);
      showTurnLabels(ggi);
      setRemoveButtonSensitive(ggi);
      break;
    case MESSAGE_REQUEST_UNDO:
      /* for now just grant the undo request
       * and send an additional request to undo his last move too... */
      if (goGameInterfaceConnectionType(ggi) == CONNECTION_SOCKET)
      {
	 if (message->owner != goGameGetId(ggi->game) &&
	     message->player == goGameGetTurn(ggi->game))
	 {
	    ggiMessageSend(ggi, True, MESSAGE_UNDO_MOVE,
			   message->x, message->y, goGameGetPlayer(ggi->game),
			   NULL);
	    /* this interface can't ask for additional undo... */
	    /*
	    ggiMessageSend(ggi, True, MESSAGE_REQUEST_UNDO,
			   message->x, message->y, goGameGetPlayer(ggi->game),
			   NULL);
			   */
	 }
      }
      break;
    case MESSAGE_NEW_GAME:
      drawBoardDifferent(ggi, copy);
      showTurnLabels(ggi);
      setRemoveButtonSensitive(ggi);
      break;
    case MESSAGE_TRANSFER_DONE:
      ggi->suppressOutput = False;
      commentAreaAppend(ggi, messtr);
      drawWholeBoard(ggi);
      highlightLastStone(ggi);
      showTurnLabels(ggi);
      setRemoveButtonSensitive(ggi);
      break;
    case MESSAGE_TRANSFER_HIST:
      if (goGameInterfaceConnectionType(ggi) == CONNECTION_SOCKET)
      {
	 /* if this came first get ready to suppress some output */
	 if (message->owner != goGameGetId(ggi->game))
	 {
	    /* clear the game */
	    goMessageSend(ggi->game, MESSAGE_NEW_GAME, 0, 0,
			   goGameGetPlayer(ggi->game), NULL);
	    commentAreaAppend(ggi, messtr);
	    ggi->suppressOutput = True;
	 }
	 else /* (message->owner == goGameGetId(ggi->game)) */
	 {
	    goMessage *transfer = NULL;
	    int recv = -1, stop = False;
	    
	    /* transfer every node of history from here */
	    for (transfer = goGameGetMessage(ggi->game);
		 transfer && !stop; transfer = goGameGetMessage(NULL))
	    {
	       goMessage *temporary = goMessageCopy(transfer);
	       
	       if (temporary)
	       {
		  temporary->override = (char)True;
		  /* printf("transferring: %s\n", goMessageStr(temporary)); */
		  sendSocket(ggi->connection, temporary, sizeof(goMessage));
		  recv = recvSocket(ggi->connection, temporary,
				    sizeof(goMessage));
		  if (recv != sizeof(goMessage) ||
		      temporary->type != MESSAGE_REQUEST_MOVE)
		  {
		     printf("Transfer stopped.\n");
		     stop = True;
		  }
		  temporary = goMessageDestroy(temporary);
	       }
	    }
	 }
      }
      break;
    case MESSAGE_COMMENT:
      commentAreaAppend(ggi, messtr);
      break;
    case MESSAGE_NOTIFY_CLIENT:
      if (message->owner != goGameGetId(ggi->game))
      {
	 sprintf(messtr, "* Opponent (%s) connected.\n",
		 message->comment);
	 commentAreaAppend(ggi, messtr);
      }
      break;
    case MESSAGE_NOTIFY_SERVER:
      if (message->owner != goGameGetId(ggi->game))
      {
	 sprintf(messtr, "* Opponent (%s) connected.\n",
		 message->comment);
	 commentAreaAppend(ggi, messtr);
	 ggiMessageSend(ggi, False, MESSAGE_NOTIFY_CLIENT,
			0, 0, goGameGetPlayer(ggi->game), getlogin());
      }
      break;
    case MESSAGE_DISCONNECT:
      if (message->owner != goGameGetId(ggi->game))
      {
	 sprintf(messtr,"* Opponent disconnected.\n");
	 commentAreaAppend(ggi, messtr);
	 /* to prevent future transmissions */
	 goGameInterfaceDisconnect(ggi);
      }
      break;
    case MESSAGE_TERRITORY_ON:
      if (message->owner == goGameGetId(ggi->game))
      {
	 Widget terrShell;
	 ggi->territory =
	    goTerritoryCreate(ggi->game, message->player);
	 terrShell = territoryPanelBuild(ggi);
      }
      break;
    case MESSAGE_TERRITORY_OFF:
      if (message->owner == goGameGetId(ggi->game) && ggi->territory)
      {
	 ggi->territory = goTerritoryDestroy(ggi->territory);
      }
      break;
   }

   copy = goGameDestroy(copy);
}

void xgoInputProc(goGameInterface *ggi, int *sourcePtr, XtInputId *id)
{
   if (goGameInterfaceConnectionType(ggi) == CONNECTION_SOCKET)
   {
      int recv;
      goMessage message;

      /* printf("in xgoInputProc()\n"); */
      recv = recvSocket(ggi->connection, (void *)&message, sizeof(goMessage));
      if (recv == sizeof(goMessage))
      {
	 /* printf("received: %s\n", goMessageStr(&message)); */
	 /* if the message was from someone else.... */
	 if (message.owner != goGameGetId(ggi->game))
	    ggiMessageApply(ggi, &message, True);
      }
   }
}

