/* xIFaceBoard.c -- handles the go interface of choice (X windows)
 * this file handles the board */

/* xmango -- copyright 1994 Edward Grigassy
 * the code included in this package may be distributed, read, and copied
 * freely BUT NOT SOLD OR INCLUDED IN OTHER SOFTWARE, for commercial or
 * public/free use, without express written consent of the author.
 */

/* include files: */
#include "xgo.h"

#include "goGameP.h"
#include "xIFaceP.h"

/* minimum number of pixels per stone space to even bother drawing the board */
#define MIN_PIXELS_PER_SPACE	7
/* minimun number of pixels per stone to draw stone 3-dimesionally */
#define MIN_PIXELS_FOR_3D	15
/* stone margin is n where empty border takes up 1/n area of each stone
 * space (the rest, of course, being filled by the stone itself) */
#define STONE_MARGIN		7

void beepGame(goGameInterface *ggi)
{
   Widget board = goGameInterfaceGetBoard(ggi);
   
   if (board)
   {
      XBell(XtDisplay(board), 50);
   }
}

void removeGroupOfStonesCB(Widget w, goGameInterface *ggi, XtPointer nothing)
{
   if (ggi && ggi->iface)
   {
      ggi->iface->nextAction = ACTION_REMOVE_GROUP;
      if (!ggi->iface->removeCursor)
      {
	 ggi->iface->removeCursor =
	    XCreateFontCursor(XtDisplay(ggi->iface->area),
			      XC_crosshair);
      }
      if (ggi->iface->removeCursor)
      {
	 XDefineCursor(XtDisplay(ggi->iface->area), XtWindow(ggi->iface->area),
		       ggi->iface->removeCursor);
      }
   }
}

void inputBoardCB(Widget w, XtPointer info, XEvent *event, Boolean *dispatch)
{
   goGameInterface *ggi = (goGameInterface *)info;

   if (ggi && ggi->iface && event && ggi->iface->drawn)
   {
      int xsz, ysz, xoff, yoff;
      int xat, yat;
      int boardx, boardy, spot;

      if (event->type == ButtonPress)
      {
	 xat = event->xbutton.x;
	 yat = event->xbutton.y;

	 getAspectInfo(ggi);
	 xsz = ggi->iface->xsz, ysz = ggi->iface->ysz;
	 xoff = ggi->iface->xoff, yoff = ggi->iface->yoff;
	 boardx = (xat-xoff+xsz/2) / xsz;
	 boardy = (yat-yoff+ysz/2) / ysz;

#if defined(NO_REPEAT_PLAYER)
	 if (goGameGetTurn(ggi->game) != goGameGetPlayer(ggi->game))
	 {
	    dialogErrorPopup(ggi->iface->area,"Move Error",
			     "It is the other player's turn.", True);
	    beepGame(ggi);
	 }	 
	 else
#endif
	    if (xat >= 0 && xat <= ggi->iface->boardWidth
	     && yat >= 0 && yat <= ggi->iface->boardHeight
	     && boardx < ggi->game->numColumns
	     && boardy < ggi->game->numRows)
	 {
	    spot = boardGetSpot(ggi->game, boardx, boardy);
	       

	    switch (ggi->iface->nextAction)
	    {
	     case ACTION_PLACE_STONE:
	       if (spot == PLAYER_NONE)
	       {
		  goMessage fillMe;
		  int reason = REASON_VALID;
   
		  goMessageFill(&fillMe, ggi->game,
				MESSAGE_PLACE_STONE,
				boardx, boardy,
				goGameGetPlayer(ggi->game), NULL);
		  /* should we beep our own terminal? True does */
		  if (goMessageValid(ggi->game, &fillMe, &reason) &&
		      reason == REASON_VALID)
		     ggiMessageApply(ggi, &fillMe, True);
		  else
		     showInvalidReason(ggi, reason);
	       }
	       else
		  beepGame(ggi);
	       break;
	       
	     case ACTION_REMOVE_GROUP:
	       /* this action has been disabled for the user */
	       if (spot != PLAYER_NONE)
	       {
		  ggiMessageSend(ggi, True, MESSAGE_CAPTURE_GROUP,
				 boardx, boardy, spot, NULL);
	       }
	       else
		  beepGame(ggi);
	       ggi->iface->nextAction = ACTION_PLACE_STONE;
	       XUndefineCursor(XtDisplay(ggi->iface->area),
			       XtWindow(ggi->iface->area));
	       break;
	    }
	 }
	 else beepGame(ggi);
      }
   }
}

/* highlights the last stone played
 * could be made to not access private elements of objects: game, history
 */

void highlightLastStone(goGameInterface *ggi)
{
   Widget theBoard = ggi->iface->area;
   GC gc = XDefaultGCOfScreen(XtScreen(theBoard));
   Display *theDisplay = XtDisplay(theBoard);
   Window theWindow = XtWindow(theBoard);
   unsigned long fore, high, low, edge, holdFore;
   int xsz = ggi->iface->xsz, ysz = ggi->iface->ysz;
   int xoff= ggi->iface->xoff, yoff = ggi->iface->yoff;
   goHistory *last = NULL, *penultimate;

   /* highlight the last stone (if this is it) */
   /* make sure there were no passes (the last stone cannot be highlighted
    * because it may have been in the last group removed) */
   if (ggi->game)
      last = goHistorySearchLast(ggi->game->historyLast, MESSAGE_PASS);
   if (last) last = NULL;
   else 
      last = goHistorySearchLast(ggi->game->historyLast, MESSAGE_PLACE_STONE);
   if (last)
   {
      int xstart, ystart, circw, circh;
      int spot = boardGetSpot(ggi->game,last->turn->x,last->turn->y);

      penultimate = goHistorySearchLast(last->prev, MESSAGE_PLACE_STONE);
      if (penultimate)
      {
	 /* this recursive call may look tacky, but we have to 
	  * "unhighlight" the penultimate stone played */
	 if (penultimate->turn->x != last->turn->x ||
	     penultimate->turn->y != last->turn->y)
	    drawStone(ggi, penultimate->turn->x, penultimate->turn->y);
      }
      if (spot == PLAYER_BLACK)
      {
	 fore = ggi->iface->bright;
      }
      else if (spot == PLAYER_WHITE)
      {
	 fore = ggi->iface->grid;
      }
      circw = xsz/3;
      circh = ysz/3;
      xstart = last->turn->x*xsz + xoff - circw/2;
      ystart = last->turn->y*ysz + yoff - circh/2;
      XSetForeground(theDisplay, gc, fore);
      XDrawArc(theDisplay, theWindow, gc,
	       xstart, ystart, circw-1, circh-1, 0, 360*64);
   }
}

void drawStone(goGameInterface *ggi, int i, int j)
{
   int xat, yat;
   Widget theBoard = ggi->iface->area;
   GC gc = XDefaultGCOfScreen(XtScreen(theBoard));
   Display *theDisplay = XtDisplay(theBoard);
   Window theWindow = XtWindow(theBoard);
   unsigned long fore, high, low, edge, holdFore;
   int xsz = ggi->iface->xsz, ysz = ggi->iface->ysz;
   int xoff= ggi->iface->xoff, yoff = ggi->iface->yoff;
   int possible3d = (xsz >= MIN_PIXELS_FOR_3D && ysz >= MIN_PIXELS_FOR_3D);
   int xmarg, ymarg, stonew, stoneh;
   int spot = boardGetSpot(ggi->game,i,j);

   if (possible3d)
   {
      xmarg = xsz/STONE_MARGIN;
      ymarg = ysz/STONE_MARGIN;
   }
   else
   {
      xmarg = xsz/(STONE_MARGIN*2);
      ymarg = ysz/(STONE_MARGIN*2);
   }
   stonew = xsz-xmarg*2-1;
   stoneh = ysz-ymarg*2-1;
   xat = i*xsz + xoff - xsz/2 + xmarg;
   yat = j*ysz + yoff - ysz/2 + ymarg;
   
   XtVaGetValues(theBoard, FOREGROUND_RESOURCE, &holdFore, NULL);
   if (spot == PLAYER_BLACK)
   {
      fore = ggi->iface->player1;
      high = ggi->iface->light;
      low = ggi->iface->dull;
      edge = ggi->iface->grid;
   }
   else if (spot == PLAYER_WHITE)
   {
      fore = ggi->iface->player2;
      high = ggi->iface->bright;
      low = ggi->iface->light;
      if (!possible3d)
	 edge = ggi->iface->grid;
      else
	 edge = ggi->iface->light;
   }
   
   if (spot != PLAYER_NONE)
   {
      if (possible3d)
      {
	 XGCValues values;
	 values.fill_style = FillOpaqueStippled;
	 /* XChangeGC(theDisplay, gc, GCFillStyle, &values); */
	 XSetForeground(theDisplay, gc, ggi->iface->shadow);
	 XFillArc(theDisplay, theWindow, gc,
		  xat-xmarg, yat+(2*ymarg), stonew-xmarg, stoneh-ymarg,
		  0*64, 360*64);
	 values.fill_style = FillSolid;
	 /* XChangeGC(theDisplay, gc, GCFillStyle, &values); */
      }
      XSetForeground(theDisplay, gc, fore);
      XFillArc(theDisplay, theWindow, gc,
	       xat, yat, stonew, stoneh,
	       0, 360*64);
      /* draw an edge around the stone */
      XSetForeground(theDisplay, gc, edge);
      XDrawArc(theDisplay, theWindow, gc,
	       xat, yat, stonew, stoneh,
	       0, 360*64);
      if (possible3d)
      {
	 int concentric;
	 
	 for (concentric = 0; concentric < 3; concentric++)
	 {
	    int xCircOffset = xsz/8+concentric,
	        yCircOffset = ysz/8+concentric;
	    
	    XSetForeground(theDisplay, gc, low);
	    XDrawArc(theDisplay, theWindow, gc,
		     xat+xCircOffset, yat+yCircOffset,
		     stonew-(xCircOffset*2),
		     stoneh-(yCircOffset*2),
		     (180+(concentric*3))*64, (45-(concentric*6))*64);
	    XSetForeground(theDisplay, gc, high);
	    XDrawArc(theDisplay, theWindow, gc, 
		     xat+xCircOffset, yat+yCircOffset,
		     stonew-(xCircOffset*2),
		     stoneh-(yCircOffset*2),
		     (concentric*3)*64, (45-(concentric*6))*64);
	 }
      }
   }
   XSetForeground(theDisplay, gc, holdFore);
}

void drawIntersection(goGameInterface *ggi, int i, int j)
{
   int xat, yat;
   Widget theBoard = ggi->iface->area;
   GC gc = XDefaultGCOfScreen(XtScreen(theBoard));
   Display *theDisplay = XtDisplay(theBoard);
   Window theWindow = XtWindow(theBoard);
   unsigned long fore, back;
   int xsz = ggi->iface->xsz, ysz = ggi->iface->ysz;
   int xoff= ggi->iface->xoff, yoff = ggi->iface->yoff;

   xat = i*xsz + xoff - xsz/2;
   yat = j*ysz + yoff - ysz/2;

   XtVaGetValues(theBoard,
		 BACKGROUND_RESOURCE, &back,
		 NULL);
   fore = ggi->iface->grid;

   if (boardGetSpot(ggi->game,i,j) == PLAYER_NONE)
   {
      int xstart, xend, ystart, yend;
     
      XSetForeground(theDisplay, gc, back);
      XFillRectangle(theDisplay, theWindow, gc, xat, yat, xsz, ysz);
      xstart = xat + (i == 0? xsz/2: 0);
      xend = xat + (i < (ggi->game->numColumns-1)? xsz: xsz/2);
      ystart = yat + (j == 0? ysz/2: 0);
      yend = yat + (j < (ggi->game->numRows-1)? ysz: ysz/2);
      XSetForeground(theDisplay, gc, fore);
      XDrawLine(theDisplay, theWindow, gc,
		xstart, yat+ysz/2, xend, yat+ysz/2);
      XDrawLine(theDisplay, theWindow, gc,
		xat+xsz/2, ystart, xat+xsz/2, yend);

      /* draw a star point if appropriate */
      if (isStarPoint(ggi->game, i, j))
      {
	 int circw, circh;
	 
	 circw = xsz/3;
	 circh = ysz/3;
	 xstart = xat + xsz/2 - circw/2;
	 ystart = yat + ysz/2 - circh/2;
	 XFillArc(theDisplay, theWindow, gc,
		  xstart, ystart, circw, circh, 0, 360*64);
      }
   }
}

void getAspectInfo(goGameInterface *ggi)
{
   Dimension uwd, uht;
   goInterface *interface = ggi->iface;
   
   XtVaGetValues(interface->area, WIDTH_RESOURCE, &uwd,
		 HEIGHT_RESOURCE, &uht, NULL); 

   interface->boardWidth = (int)uwd;
   interface->boardHeight = (int)uht;
   interface->xsz = interface->boardWidth / (ggi->game->numColumns+1);
   interface->ysz = interface->boardHeight / (ggi->game->numRows+1);
   interface->xoff = interface->xsz * 2 / 3; 
   /* + (interface->boardWidth - interface->xsz * ggi->game->numColumns)/2; */
   interface->yoff = interface->ysz * 2 / 3;
   /* + (interface->boardHeight - interface->ysz * ggi->game->numRows)/2; */
}

void drawRowLabel(goGameInterface *ggi,
		  Display *display, Window window, GC gc,
		  int row, int x, int y)
{
   char label[10];

   sprintf(label, "%d", ggi->game->numRows - row);
   XDrawString(display, window, gc, x, y, label, strlen(label));
}

void drawColumnLabel(goGameInterface *ggi,
		     Display *display, Window window, GC gc,
		     int column, int x, int y)
{
   char label[10], colChar;
   
   colChar = 'A'+column;
   if (colChar >= 'I') colChar++;
   sprintf(label, "%c", colChar);
   XDrawString(display, window, gc, x, y, label, strlen(label));
}

void drawWholeBoard(goGameInterface *ggi)
{
   /* draw the board */
   Widget theBoard;
   int i, j, from, to, lineAt, midRow, midColumn, starOffset;
   Display *theDisplay;
   Window theWindow;
   GC gc;
   unsigned long fore;
   int xsz, ysz;
   int xoff, yoff;

   xsz = ggi->iface->xsz, ysz = ggi->iface->ysz;
   xoff = ggi->iface->xoff, yoff = ggi->iface->yoff;

   theBoard = ggi->iface->area;
   theDisplay = XtDisplay(theBoard);
   theWindow = XtWindow(theBoard);
   gc = XDefaultGCOfScreen(XtScreen(theBoard));
   
   /* draw the grid */
   fore = ggi->iface->grid;
   XSetForeground(theDisplay, gc, fore);
   XClearWindow(theDisplay, theWindow);
   /* draw vertical lines */
   for (i = 0; i < ggi->game->numColumns; i++)
   {
      from = yoff;
      to = (ggi->game->numColumns-1)*ysz + yoff;
      lineAt = (i*xsz + xoff);
      XDrawLine(theDisplay, theWindow, gc,
		lineAt, from, lineAt, to);
      drawColumnLabel(ggi, theDisplay, theWindow, gc, i,
		      lineAt-(xoff/4), to+ysz);
   }
   /* draw horizontal lines */
   for (j = 0; j < ggi->game->numRows; j++)
   {
      from = xoff;
      to = (ggi->game->numRows-1)*xsz + xoff;
      lineAt = (j*ysz + yoff);
      XDrawLine(theDisplay, theWindow, gc,
		from, lineAt, to, lineAt);
      drawRowLabel(ggi, theDisplay, theWindow, gc, j,
		   to+xoff, lineAt+(yoff/2));
   }

   /* draw the star points */
   if (ggi->game->starPoint)
   {
      midRow = (ggi->game->numRows)/2;
      midColumn = (ggi->game->numColumns)/2;
      starOffset = ggi->game->starPoint-1;
      drawIntersection(ggi, starOffset, starOffset);
      drawIntersection(ggi, starOffset, midRow);
      drawIntersection(ggi, starOffset, ggi->game->numRows-starOffset-1);
      drawIntersection(ggi, midColumn, starOffset);
      drawIntersection(ggi, midColumn, midRow);
      drawIntersection(ggi, midColumn, ggi->game->numRows-starOffset-1);
      drawIntersection(ggi, ggi->game->numColumns-starOffset-1, starOffset);
      drawIntersection(ggi, ggi->game->numColumns-starOffset-1, midRow);
      drawIntersection(ggi, ggi->game->numColumns-starOffset-1,
		       ggi->game->numRows-starOffset-1);
   }
   
   /* draw the pieces */
   for (i = 0; i < ggi->game->numRows; i++)
   {
      for (j = 0; j < ggi->game->numColumns; j++)
      {
	 int spot = boardGetSpot(ggi->game, i, j);
	 if (spot != PLAYER_NONE)
	    drawStone(ggi, i, j);
      }
   }
   ggi->iface->drawn = True;
}

void showEventInfo(Widget w, goGame *theGame, XEvent *event, Boolean *contOn);


void redrawBoardCB(Widget w, XtPointer info, XEvent *event, Boolean *dispatch)
{
   goGameInterface *ggi = (goGameInterface *)info;
   int ignoredEvent = False;

   if (ggi && XtIsRealized(ggi->iface->area))
   {
      getAspectInfo(ggi);
      
      /* make sure the drawing area is big enough */
      if (ggi->iface->boardWidth >=
	  (ggi->game->numColumns * MIN_PIXELS_PER_SPACE) &&
	  ggi->iface->boardHeight >=
	  (ggi->game->numRows * MIN_PIXELS_PER_SPACE))
      {
	 int force = (event->type == Expose && event->xexpose.count == 0) ||
	    event->type == ConfigureNotify;
#if defined(DEBUG_EVENTS)
	 showEventInfo(w, ggi->game, event, NULL);
#endif

	 if (force && !ggi->suppressOutput)
	 {
	    drawWholeBoard(ggi);
	    highlightLastStone(ggi);
	 }
      }
      else /* just clear it out */
	 XClearWindow(XtDisplay(ggi->iface->area),
		      XtWindow(ggi->iface->area));
   }
}

/* only draw stones/intersections where they are different
 * (i.e. where ggi->game != orig)
 */

void drawBoardDifferent(goGameInterface *ggi, goGame *orig)
{
   if (ggi && orig)
   {
      int i, j, newSpot, oldSpot;
      
      for (j = 0; j < orig->numRows; j++)
	 for (i = 0; i < orig->numColumns; i++)
	 {
	    newSpot = boardGetSpot(ggi->game, i, j);
	    oldSpot = boardGetSpot(orig, i, j);
	    if (newSpot != oldSpot)
	    {
	       if (newSpot == PLAYER_NONE)
		  drawIntersection(ggi, i, j);
	       else
		  drawStone(ggi, i, j);
	    }
	 }
   }
}

#if defined(DEBUG_EVENTS)

#include <X11/IntrinsicP.h>
#include <X11/CoreP.h>

static char *eventTypes[] = {
  "Event0?",
  "Event1?",
  "KeyPress",
  "KeyRelease",
  "ButtonPress",
  "ButtonRelease",
  "MotionNotify",
  "EnterNotify",
  "LeaveNotify",
  "FocusIn",
  "FocusOut",
  "KeymapNotify",
  "Expose",
  "GraphicsExpose",
  "NoExpose",
  "VisibilityNotify",
  "CreateNotify",
  "DestroyNotify",
  "UnmapNotify",
  "MapNotify",
  "MapRequest",
  "ReparentNotify",
  "ConfigureNotify",
  "ConfigureRequest",
  "GravityNotify",
  "ResizeRequest",
  "CirculateNotify",
  "CirculateRequest",
  "PropertyNotify",
  "SelectionClear",
  "SelectionRequest",
  "SelectionNotify",
  "ColormapNotify",
  "ClientMessage",
  "MappingNotify",
  "LASTEvent",
  NULL
};

void showEventInfo(Widget w, goGame *theGame, XEvent *event, Boolean *contOn)
{
   int wid = 0, hgt = 0;

   printf("%x %s (%s) Widget w, h: %d, %d; ",
	  event->type, eventTypes[event->type],
	  w->core.name, w->core.width, w->core.height);
   if (event->type == ConfigureNotify)
   {
      wid = event->xconfigure.width;
      hgt = event->xconfigure.height;
      printf("Event w, h: %d, %d", wid, hgt);
   }
   else if (event->type == Expose)
   {
      wid = event->xexpose.width;
      hgt = event->xexpose.height;
      printf("Event w, h: %d, %d; count: %d", wid, hgt,
	     event->xexpose.count);
   }
   printf("\n");
}

#endif

